require "spec_helper"

describe Matador::Fetcher do
  before(:all) do
    @redis = Redis.new(host: "127.0.0.1", port: "6379")
    Matador.cache_store = @redis 
    @redis.del("test_key1")
    @redis.del("matador:nominal_ttl:test_key1")
    @redis.del("matador:rebuilding:test_key1")
  end

  context "#fetch" do
    after(:each) do
      @redis.del("test_key1")
      @redis.del("matador:nominal_ttl:test_key1")
      @redis.del("matador:rebuilding:test_key1")
    end

    it "#fetch gets the cache if it is still valid" do
      Matador.fetch("test_key1", :ttl => 60, :therd_ttl => 10) do
        "foobar"
      end
      # set value to foobar

      value = Matador.fetch("test_key1", :ttl => 60, :therd_ttl => 10) do
        "foobaz"
      end
      # cache should still be active so value sould be foobar, not foobaz

      expect(value == "foobar").to be true
    end
   
    it "#fetch kicks off a single request to rebuild cache after it expires" do
      Matador::Fetcher.any_instance.stub(:rebuild_cache_with_first_request).and_return("first_of_the_herd")

      Matador.fetch("test_key1", :ttl => 1, :therd_ttl => 5) do
        "foobar"
      end

      sleep 2

      value = Matador.fetch("test_key1", :ttl => 1, :therd_ttl => 5) do
        "foobaz"
      end

      expect(value).to eq("first_of_the_herd")
    end

    it "#fetch returns stale_value after first post-expiration request went through to rebuild cache" do
      Matador::Fetcher.any_instance.stub(:locked_for_rebuilding_cache?).and_return(true)
      Matador::Fetcher.any_instance.stub(:valid_cache?).and_return(false)
      Matador::Fetcher.any_instance.stub(:current_cache).and_return("stale cache")

      value = Matador.fetch("test_key1", :ttl => 1, :therd_ttl => 5) do
        "foobar"
      end
      
      expect(value).to eq("stale cache")
    end

    it "no matter how many requests come in after the first request, they do not rebuild the cache" do
      @redis.set("test_key1", "stale_cache")

      Thread.new do
        Matador.fetch("test_key1", :ttl => 100, :therd_ttl => 12) do
          sleep 10 
          "foobar"
        end
      end

      sleep 2

      value1 = Matador.fetch("test_key1", :ttl => 100, :therd_ttl => 10) do
        "foobar"
      end

      sleep 2 

      value2 = Matador.fetch("test_key1", :ttl => 100, :therd_ttl => 10) do
        "foobar"
      end

      sleep 2

      value3 = Matador.fetch("test_key1", :ttl => 100, :therd_ttl => 10) do
        "foobar"
      end

      sleep 8

      value4 = Matador.fetch("test_key1", :ttl => 100, :therd_ttl => 10) do
        "foobar"
      end
   
      expect(value1).to eq("stale_cache")
      expect(value2).to eq("stale_cache")
      expect(value3).to eq("stale_cache")
      expect(value4).to eq("foobar")
    end

    context "when therd_ttl is not long enough" do
      after(:each) do
        @redis.del("test_key1")
        @redis.del("matador:nominal_ttl:test_key1")
        @redis.del("matador:rebuilding:test_key1")
      end

      it "returns the computation once while cache is building in the background" do
        @redis.set("test_key1", "stale_cache")

        Thread.new do
          Matador.fetch("test_key1", :ttl => 100, :therd_ttl => 5) do
            sleep 10 
            "foobar"
          end
        end

        value = Matador.fetch("test_key1", :ttl => 100, :therd_ttl => 10) do
          "foobaz"
        end

        expect(value).to eq("foobaz")
      end

      it "rebuilds the cache with first request if that has not happened already" do
        @redis.set("test_key1", "stale_cache")

        Thread.new do
          Matador.fetch("test_key1", :ttl => 100, :therd_ttl => 3) do
            sleep 100
            "foobar"
          end
        end

        sleep 5

        Matador.fetch("test_key1", :ttl => 100, :therd_ttl => 10) do
          "foobaz"
        end

        value = @redis.get("test_key1")

        expect(value).to eq("foobaz")
      end
    end
  end
end

